#!/usr/bin/perl
#
# Author: Rafael A. Nafikov email <rnafscience@gmail.com>
#
#The ADMIXFRQ Perl script computes marker allele frequencies in admixed populations using any combination of four different admixture models described in Nafikov et al., 2018. 
#
#
#
#

use strict;
use warnings;


########### controls ####################################################################

# $global_adm_comp = 0 or 1 (0 = use supplied global admixture proportions, 1 = compute admixture proportions from data)

# $adm_printout = 0 or 1 (0 = do not print out or 1 = print out admixture proportions)
# $fbcw_adm_where = 0 or 1 (0 = use $fbcw_adm_ref or 1 = use $fbcw_adm_b_ref for adm prinout)

# $control_g = 0 (nothing) or 1 for Global model
# $control_fbgw = 0 (nothing) or 1 for FBGW model
# $control_fbcw = 0 (nothing) or 1 for FBCW model
# $control_l = 0 (nothing) or 1 for Local model

# $control_1 = 0 (nothing) or 1 for computing allele frequencies for geno file
# $control_2 = 0 (nothing) or 1 for computing allele frequencies not for geno file

# $control_geno_extra = 0 (nothing) or 1 for printing out geno file frequencies into seporate file

# $comp_type = 0 if frequencies for geno files are needed; 1 = if frequencies not for geno files are needed (FBGW, FBCW, Local); 2 = if admixture proportions for Local model are needed; 3 = if admixture proportions for Global, FBGW, or FBCW model are needed; 4 = directory for error log file; 5 = if frequencies not for geno files are needed (Global)
##########################################################################################


	main ();

sub main {

	$| = 1;
	my $parameters_ref = read_par_file ();
	my %parameters = %$parameters_ref;
	script_info ($parameters_ref);

	parameters_check ($parameters_ref);


	my @chrs;
	for my $key1 (sort keys %parameters) {
	if (defined $parameters{$key1}{'CHROMOSOMES'}) {
	for my $key2 (sort {$a <=> $b} keys %{$parameters{$key1}{'CHROMOSOMES'}}) { 
	if ($parameters{$key1}{'CHROMOSOMES'}{$key2} =~ m/\.\./) {
	my @temp1 = split (/\.\./, $parameters{$key1}{'CHROMOSOMES'}{$key2});
	my @temp2 = ($temp1[0]..$temp1[1]);
	push @chrs, @temp2;
	} else {
	my @temp1 = split (/\.\./, $parameters{$key1}{'CHROMOSOMES'}{$key2});
	push @chrs, @temp1;
	}
	}
	}
	}
	error ("Error: chromosome numbers were not specified!") if (! @chrs);

	my @families;
	for my $key1 (sort keys %parameters) {
	if (defined $parameters{$key1}{'FAMILYNAMES'}) {
	for my $key2 (sort {$a <=> $b} keys %{$parameters{$key1}{'FAMILYNAMES'}}) { 
	push @families, $parameters{$key1}{'FAMILYNAMES'}{$key2};
	}
	}
	}
	error ("Error: family names were not provided!") if (! @families);

	my @models;
	for my $key1 (sort keys %parameters) {
	if (defined $parameters{$key1}{'ANCESTRYMODEL'}) {
	for my $key2 (sort {$a <=> $b} keys %{$parameters{$key1}{'ANCESTRYMODEL'}}) { 
	push @models, $parameters{$key1}{'ANCESTRYMODEL'}{$key2};
	}
	}
	}
	error ("Error: ancestry models were not specified!") if (! @models);

	my $global_adm_comp = 0;
	for my $key1 (sort keys %parameters) {
	for my $key2 (sort keys %{$parameters{$key1}}) {
	if ($key2 eq 'COMPUTEGLOBALANCESTRYMODELPROPORTIONS') {
	if (uc $parameters{$key1}{'COMPUTEGLOBALANCESTRYMODELPROPORTIONS'} eq 'YES') {
	$global_adm_comp = 1;
	}
	}
	}
	}
	my $global_adm_comp_ref = \$global_adm_comp;

	my $adm_printout = 0; # 0 = do not export admixture proportions; 1 = export admixture proportions
	for my $key1 (sort keys %parameters) {
	for my $key2 (sort keys %{$parameters{$key1}}) {
	if ($key2 eq 'ADMIXTUREPROPORTIONSPRINTOUT') {
	if (uc $parameters{$key1}{'ADMIXTUREPROPORTIONSPRINTOUT'} eq 'YES') {
	$adm_printout = 1;
	}
	}
	}
	}
	my $adm_printout_ref = \$adm_printout;

	my $local_adm_extrapolation = 0; # 0 = extrapolate local adms or 1 = do not extrapolate
	for my $key1 (sort keys %parameters) {
	for my $key2 (sort keys %{$parameters{$key1}}) {
	if ($key2 eq 'EXTRAPOLATIONOFLOCALADMIXTUREPROPORTIONS') {
	if (uc $parameters{$key1}{'EXTRAPOLATIONOFLOCALADMIXTUREPROPORTIONS'} eq 'NO') {
	$local_adm_extrapolation = 1;
	}
	}
	}
	}
	my $local_adm_extrapolation_ref = \$local_adm_extrapolation;


	my $control_g = 0;
	my $control_fbgw = 0;
	my $control_fbcw = 0;
	my $control_l = 0;

	for my $el (@models) {
	$el = uc $el;
	if ($el eq 'GLOBAL') {
	$control_g = 1;
	}
	elsif ($el eq 'FBGW') {
	$control_fbgw = 1;
	}
	elsif ($el eq 'FBCW') {
	$control_fbcw = 1;
	}
	elsif ($el eq 'LOCAL') {
	$control_l = 1;
	}
	}


	my $control_1 = 0; # option for geno file
	my $control_geno_extra = 0; # option to ouput geno file frequencies seporately
	for my $key1 (sort keys %parameters) {
	if (defined $parameters{$key1}{'GENOFILEFORGL_AUTO'}) {
	if (exists $parameters{$key1}{'FILEPATH'}{1}) {
	$control_1 = 1;
	} else {
	error ("Error: location of geno files was not provided! Please specify the location or remove 'GENO FILE FOR GL_AUTO' option from parameter file.");
	}
	if (uc $parameters{$key1}{'DETAILEDALLELEFREQUENCYFILE'}{1} eq "YES") {
	$control_geno_extra = 1;
	}
	}
	}

	my $control_2 = 0; # option for just modifying allele frequencies
	for my $key1 (sort keys %parameters) {
	if (defined $parameters{$key1}{'MODIFYALLELEFREQUENCIES'}) {
	if ($parameters{$key1}{'FILEPATH'}{1}) {
	$control_2 = 1;
	} else {
	error ("Error: location of file with base pair positions for which allele frequencies were requested to be modified was not provided! Please specify the location or remove 'MODIFY ALLELE FREQUENCIES' option from parameter file.");
	}
	}
	}


	my $fbgw_adm_ref;
	my $global_adm_ref; 
	my $fbcw_adm_b_ref;
	my $fbcw_adm_where = 0;


	if ($control_fbgw == 1 or $control_g = 1) { #

	my $average_type = 0;
	for my $key1 (sort keys %parameters) {
	if (defined $parameters{$key1}{'AVERAGETYPEFORGLOBALANDFBGW'}) {

	if (uc $parameters{$key1}{'AVERAGETYPEFORGLOBALANDFBGW'}{1} eq 'WEIGHTED') {
	$average_type = 1;
	}

	elsif (uc $parameters{$key1}{'AVERAGETYPEFORGLOBALANDFBGW'}{1} eq 'NORMAL') {
	$average_type = 2;
	}

	}
	}


	my $average_type_ref = \$average_type;
	my $pop_adms_genome_ref = read_file_admixtures_genome ($parameters_ref);

	($fbgw_adm_ref, $global_adm_ref) = adms_global_fbgw ($average_type_ref, $pop_adms_genome_ref, $global_adm_comp_ref) if ($global_adm_comp == 1);

	$global_adm_ref = adm_global_supplied ($parameters_ref) if ($global_adm_comp == 0);

	$fbgw_adm_ref = adms_global_fbgw ($average_type_ref, $pop_adms_genome_ref, $global_adm_comp_ref) if ($global_adm_comp == 0);

	print_fbgw_adm_file ($fbgw_adm_ref, $parameters_ref) if ($control_fbgw == 1 && $adm_printout == 1);

	print_global_adm_file ($global_adm_ref, $parameters_ref) if ($control_g == 1 && $adm_printout == 1);

	$fbcw_adm_b_ref = adms_fbcw_b ($pop_adms_genome_ref) if ($adm_printout == 1 && $control_fbcw == 1); 
	print_fbcw_adm_file ($fbcw_adm_b_ref, $parameters_ref, 0, 0) if ($adm_printout == 1 && $control_fbcw == 1);
	$fbcw_adm_where = 1 if ($adm_printout == 1 && $control_fbcw == 1);



	} #



	for my $chr (@chrs) {
	my $chr_ref = \$chr;

###
	my $pop_frqs_ref = read_file_subpopulation ($parameters_ref, $chr_ref);
	my $pop_adms_ref = read_file_admixtures_chr ($parameters_ref, $chr_ref);
	my $sample_frqs_ref = read_file_sample_allele_frqs ($parameters_ref, $chr_ref) if ($control_2 == 1);
###


	my ($bp_pos_for_frq_ref, $frq_template_ref, $subpop_frq_template_ref, $global_frqs_ref);
	if ($control_2 == 1) { # just bp
	$bp_pos_for_frq_ref = read_bp_modify_frq ($parameters_ref, $chr_ref) if ($control_2 == 1);
	($frq_template_ref, $subpop_frq_template_ref) = frq_template ($parameters_ref, $sample_frqs_ref, $pop_frqs_ref, $bp_pos_for_frq_ref, $chr_ref) if ($control_2 == 1);

	$global_frqs_ref = global_adm_frqs ($frq_template_ref, $chr_ref, $subpop_frq_template_ref, $global_adm_ref) if ($control_g == 1 && $control_2 == 1);
	print_frqs ($parameters_ref, $global_frqs_ref, $chr_ref, 0, 1) if ($control_g == 1 && $control_2 == 1);

	}






	for my $family (@families) {
	my $family_ref = \$family;




######
	my $fbcw_adms_ref = adms_fbcw ($pop_adms_ref) if ($control_fbcw == 1);
	remove_old_fbcw_adm_file ($parameters_ref, $chr_ref, $family_ref) if ($adm_printout == 1 && $control_fbcw == 1 && $fbcw_adm_where == 0);
	print_fbcw_adm_file ($fbcw_adms_ref, $parameters_ref, $chr_ref, $family_ref) if ($adm_printout == 1 && $control_fbcw == 1 && $fbcw_adm_where == 0);

	my ($local_adms_ref, $subpop_local_ref) = adms_local ($parameters_ref, $pop_adms_ref, $bp_pos_for_frq_ref, $chr_ref, $family_ref, $local_adm_extrapolation_ref) if ($control_l == 1);
	print_local_adm_file ($local_adms_ref, $subpop_local_ref, $parameters_ref, $chr_ref, $family_ref) if ($adm_printout == 1 && $control_l == 1);
#####



	if ($control_2 == 1) { # just bp

	my $fbgw_frqs_ref = fbgw_frqs ($frq_template_ref, $subpop_frq_template_ref, $fbgw_adm_ref,  $family_ref, $chr_ref) if ($control_fbgw == 1);
	print_frqs ($parameters_ref, $fbgw_frqs_ref, $chr_ref, $family_ref, 2) if ($control_fbgw == 1);

	my $fbcw_frqs_ref = fbcw_frqs ($frq_template_ref, $subpop_frq_template_ref, $fbcw_adms_ref, $family_ref, $chr_ref) if ($control_fbcw == 1);
	print_frqs ($parameters_ref, $fbcw_frqs_ref, $chr_ref, $family_ref, 3) if ($control_fbcw == 1);

	my $local_frqs_ref = local_frqs ($frq_template_ref, $subpop_frq_template_ref, $local_adms_ref,  $family_ref, $chr_ref, $subpop_local_ref, 0) if ($control_l == 1);
	print_frqs ($parameters_ref, $local_frqs_ref, $chr_ref, $family_ref, 4) if ($control_l == 1);

	}







	if (($control_1 == 1)) { # geno
	my ($marker_cm_pos_ref, $geno_bp_ref, $geno_genotypes_ref, $geno_line_1_ref) = read_geno_file_info ($parameters_ref, $chr_ref, $family_ref);

	my ($frq_template_geno_ref, $subpop_frq_template_geno_ref) = frq_template ($parameters_ref, $sample_frqs_ref, $pop_frqs_ref, $geno_bp_ref, $chr_ref);

	my $global_frqs_geno_ref = global_adm_frqs ($frq_template_geno_ref, $chr_ref, $subpop_frq_template_geno_ref, $global_adm_ref) if ($control_g == 1);

	print_geno_file ($geno_genotypes_ref, $geno_line_1_ref, $global_frqs_geno_ref, $parameters_ref, $chr_ref, $family_ref, 1) if ($control_g == 1); # 1 = Global; 2 = FBGW; 3 = FBCW; 4 = Local
	print_geno_file_frqs ($marker_cm_pos_ref, $global_frqs_geno_ref, $parameters_ref, $chr_ref, $family_ref, 1) if ($control_geno_extra == 1 && $control_g == 1);

	my $fbgw_frqs_geno_ref = fbgw_frqs ($frq_template_geno_ref, $subpop_frq_template_geno_ref, $fbgw_adm_ref,  $family_ref, $chr_ref) if ($control_fbgw == 1);
	print_geno_file ($geno_genotypes_ref, $geno_line_1_ref, $fbgw_frqs_geno_ref, $parameters_ref, $chr_ref, $family_ref, 2) if ($control_fbgw == 1); # 1 = Global; 2 = FBGW; 3 = FBCW; 4 = Local
	print_geno_file_frqs ($marker_cm_pos_ref, $fbgw_frqs_geno_ref, $parameters_ref, $chr_ref, $family_ref, 2) if ($control_geno_extra == 1 && $control_fbgw == 1);

	my $fbcw_frqs_geno_ref = fbcw_frqs ($frq_template_geno_ref, $subpop_frq_template_geno_ref, $fbcw_adms_ref, $family_ref, $chr_ref) if ($control_fbcw == 1);
	print_geno_file ($geno_genotypes_ref, $geno_line_1_ref, $fbcw_frqs_geno_ref, $parameters_ref, $chr_ref, $family_ref, 3) if ($control_fbcw == 1); # 1 = Global; 2 = FBGW; 3 = FBCW; 4 = Local
	print_geno_file_frqs ($marker_cm_pos_ref, $fbcw_frqs_geno_ref, $parameters_ref, $chr_ref, $family_ref, 3) if ($control_geno_extra == 1 && $control_fbcw == 1);

	my ($local_frqs_geno_ref, $local_adms_geno_ref);
	$local_frqs_geno_ref = local_frqs ($frq_template_geno_ref, $subpop_frq_template_geno_ref, $local_adms_ref,  $family_ref, $chr_ref, $subpop_local_ref, 0) if ($control_l == 1 && $adm_printout == 0);
	($local_frqs_geno_ref, $local_adms_geno_ref) = local_frqs ($frq_template_geno_ref, $subpop_frq_template_geno_ref, $local_adms_ref,  $family_ref, $chr_ref, $subpop_local_ref, 1) if ($control_l == 1 && $adm_printout == 1);
	print_geno_file ($geno_genotypes_ref, $geno_line_1_ref, $local_frqs_geno_ref, $parameters_ref, $chr_ref, $family_ref, 4) if ($control_l == 1); # 1 = Global; 2 = FBGW; 3 = FBCW; 4 = Local
	print_geno_file_frqs ($marker_cm_pos_ref, $local_frqs_geno_ref, $parameters_ref, $chr_ref, $family_ref, 4) if ($control_geno_extra == 1 && $control_l == 1);


	print_local_adm_geno_file ($marker_cm_pos_ref, $local_adms_geno_ref, $parameters_ref, $chr_ref, $family_ref) if ($adm_printout == 1 && $control_l == 1);

	}



	} # end of family loop






	} # end of chr loop


	my $date_time = localtime;
	print LOGFILE "	Analysis finished at: $date_time\n";
	close LOGFILE;


}


sub warning
{
	my (@warning) = @_;
	if ( scalar @warning ) {
	print "	@warning\n\n";
	print LOGFILE "	@warning\n\n"; 
	}
}


sub error
{
	my (@errors) = @_;
	if ( scalar @errors ) {
	print "	@errors\n\n";
	print LOGFILE "	@errors\n\n"; 
	}
	my $date_time = localtime;
	print "	Program was terminated at: $date_time\n\n";
	print LOGFILE "	Program was terminated at: $date_time\n\n";
	exit 1;
}


sub error_2
{
	my (@errors) = @_;
	if ( scalar @errors ) { warn "@errors\n"; }
	my $date_time = localtime;
	print "	Analysis started at: $date_time\n\n";
	print "  \@----------------------------------------------------------\@\n",
	 "  |    ADMIXFRQ!  |     v1.00      |   19/Jan/2018           |\n",
	 "  |----------------------------------------------------------|\n",
	 "  | (C) 2018 Rafael Nafikov, GNU General Public License, v3  |\n",
	 "  |----------------------------------------------------------|\n",
	 "  | For documentation, citation & bug-reporting instructions:|\n",
	 "  |        https://github.com/RafPrograms/ADMIXFRQ           |\n",
	 "  \@----------------------------------------------------------\@\n",
	 "\n";
	$date_time = localtime;
	print "	Program was terminated at: $date_time\n\n";
	exit 1;
}


sub read_par_file {
	my %options = ();
	while (defined(my $arg = shift(@ARGV))) {
	open(PARAM, $arg) or error ("Error: cannot open parameter file $arg!");
	while (defined (my $line1 = <PARAM>)) {
	chomp $line1;
	$line1 =~ s/^\s+//g;
	$line1 =~ s/\s+$//g;
	next if (! defined $line1);
	next if ($line1 =~ m/^#/);
	my @temp0 = split (/#/, $line1);
	goto LABEL1 if (! defined $temp0[0]);

	if ($temp0[0] =~ m/:/) {
	my @temp1 = split (/:/, $temp0[0]);
	$temp1[0] = uc $temp1[0];
	$temp1[0] =~ s/\s+//g;
	$temp1[1] =~ s/^\s+//g if ($temp1[1]);
	$temp1[1] =~ s/\s+$//g if ($temp1[1]);
	$options{$.}{$temp1[0]} = 1 if (defined $temp1[1] && ($temp1[0] ne 'ADMIXTUREPROPORTIONSPRINTOUT' && $temp1[0] ne 'EXTRAPOLATIONOFLOCALADMIXTUREPROPORTIONS' && $temp1[0] ne 'COMPUTEGLOBALANCESTRYMODELPROPORTIONS'));
	$options{$.}{$temp1[0]} = $temp1[1] if (defined $temp1[1] && ($temp1[0] eq 'ADMIXTUREPROPORTIONSPRINTOUT' || $temp1[0] eq 'EXTRAPOLATIONOFLOCALADMIXTUREPROPORTIONS' || $temp1[0] eq 'COMPUTEGLOBALANCESTRYMODELPROPORTIONS'));
	my @temp2 = split (/\|/, $temp1[1]);
	for my $el (@temp2)	{
	my @temp3 = split (/=/, $el);
	goto LABEL1 if (!exists $temp3[1]);
	$temp3[1] =~ s/^\s+//g;
	$temp3[1] =~ s/\s+$//g;
	if ($temp3[1] =~ m/^\[/ && $temp3[1] =~ m/\]$/) {
	$temp3[1] =~ s/^\[//g;
	$temp3[1] =~ s/\]$//g;
	my @temp4 = split (/\s+/, $temp3[1]);
	my $count = 0;
	for my $el2 (@temp4) {
	$count++;
	$temp3[0] =~ s/^\s+//g;
	$temp3[0] =~ s/\s+$//g;
	$temp3[0] = uc $temp3[0];
	$temp3[0] =~ s/\s+//g;
	$options{$.}{$temp3[0]}{$count} = $el2;
	}
	} else {
	my @temp4 = split (/\s+/, $temp3[1]);
	my $count = 0;
	for my $el2 (@temp4) {
	$count++;
	$temp3[0] =~ s/^\s+//g;
	$temp3[0] =~ s/\s+$//g;
	$temp3[0] = uc $temp3[0];
	$temp3[0] =~ s/\s+//g;
	$options{$.}{$temp3[0]}{$count} = $el2;
	}
	}
	}
	} else {

	my @temp2 = split (/\|/, $temp0[0]);
	for my $el (@temp2)	{
	my @temp3 = split (/=/, $el);
	goto LABEL1 if (!exists $temp3[1]);
	$temp3[1] =~ s/^\s+//g;
	$temp3[1] =~ s/\s+$//g;
	if ($temp3[1] =~ m/^\[/ && $temp3[1] =~ m/\]$/) {
	$temp3[1] =~ s/^\[//g;
	$temp3[1] =~ s/\]$//g;
	my @temp4 = split (/\s+/, $temp3[1]);
	my $count = 0;
	for my $el2 (@temp4) {
	$count++;
	$temp3[0] =~ s/^\s+//g;
	$temp3[0] =~ s/\s+$//g;
	$temp3[0] = uc $temp3[0];
	$temp3[0] =~ s/\s+//g;
	$options{$.}{$temp3[0]}{$count} = $el2;
	}
	} else {
	my @temp4 = split (/\s+/, $temp3[1]);
	my $count = 0;
	for my $el2 (@temp4) {
	$count++;
	$temp3[0] =~ s/^\s+//g;
	$temp3[0] =~ s/\s+$//g;
	$temp3[0] = uc $temp3[0];
	$temp3[0] =~ s/\s+//g;
	$options{$.}{$temp3[0]}{$count} = $el2;
	}
	}
	}


	}

	LABEL1:
	}
	}

	return \%options;
}


sub parameters_check {

	my ($parameters_ref) = $_[0];
	my %parameters = %$parameters_ref;





}


sub read_file_subpopulation {

	my ($parameters_ref, $chr_ref) = @_;
	my $chr = $$chr_ref;
	my %parameters = %$parameters_ref;

	my %ref_frqs = ();
	for my $key1 (sort keys %parameters) { # line numbers in parameter file
	if (defined $parameters{$key1}{'SUBPOPULATIONNAME'}) {
	my $subpop_name = $parameters{$key1}{'SUBPOPULATIONNAME'}{1};
	error ("Error: subpopulation name for 'LOCAL ADMIXTURE PER FAMILY FOR SUBPOPULATION' option was not provided!") if (!defined $subpop_name);
	my $file_path = $parameters{$key1}{'FILEPATH'}{1};

	$file_path =~ s/^"//g;
	$file_path =~ s/"$//g;
	$file_path =~ s/^'//g;
	$file_path =~ s/'$//g;
	$file_path =~ s/\*/$chr/g;
	error ("Error: file path for 'LOCAL ADMIXTURE PER FAMILY FOR SUBPOPULATION' option was not provided!") if (!defined $file_path);
	my $header = $parameters{$key1}{'HEADER'}{1};
	my $bp_position = $parameters{$key1}{'BP_POSITION'}{1};
	my $rs_number = $parameters{$key1}{'RS_NUMBER'}{1};
	my $ref_allele = $parameters{$key1}{'REF_ALLELE'}{1};
	my $alt_allele = $parameters{$key1}{'ALT_ALLELE'}{1};
	my $ref_allele_frq = $parameters{$key1}{'REF_ALLELE_FRQ'}{1};
	my $alt_allele_frq = $parameters{$key1}{'ALT_ALLELE_FRQ'}{1};

	open(FILE1, $file_path) or error ("Error: cannot open file $file_path containing reference subpopulation allele frequencies!");
	while (defined (my $line1 = <FILE1>)) {
	chomp $line1;
	$line1 =~ s/^\s+//g;
	$line1 =~ s/\s+$//g;
	next if ($header eq 'T' && $. == 1);
	my @temp1 = split (/\s+/, $line1);

	$ref_frqs{$chr}{$temp1[$bp_position-1]}{$subpop_name}{'RS_NUMBER'} = $temp1[$rs_number-1];
	$ref_frqs{$chr}{$temp1[$bp_position-1]}{$subpop_name}{'REF_ALLELE'} = $temp1[$ref_allele-1];
	$ref_frqs{$chr}{$temp1[$bp_position-1]}{$subpop_name}{'REF_ALLELE_FRQ'} = $temp1[$ref_allele_frq-1];
	$ref_frqs{$chr}{$temp1[$bp_position-1]}{$subpop_name}{'ALT_ALLELE'} = $temp1[$alt_allele-1];
	$ref_frqs{$chr}{$temp1[$bp_position-1]}{$subpop_name}{'ALT_ALLELE_FRQ'} = $temp1[$alt_allele_frq-1];

	}
	close FILE1;
	}
	}

	return \%ref_frqs;
}


sub read_file_admixtures_chr {

	my ($parameters_ref, $chr_ref) = @_;
	my $chr = $$chr_ref;
	my %parameters = %$parameters_ref;

	my %adms = ();
	my %rs_to_bp = ();

	for my $key1 (sort keys %parameters) { # line numbers in parameter file
	if (defined $parameters{$key1}{'EXTRAINFORMATIONFORADMIXTUREFILES'}) {
	my $file_path = $parameters{$key1}{'FILEPATH'}{1};
	$file_path =~ s/^"//g;
	$file_path =~ s/"$//g;
	$file_path =~ s/^'//g;
	$file_path =~ s/'$//g;
	$file_path =~ s/\*/$chr/g;
	my $header = $parameters{$key1}{'HEADER'}{1};
	my $bp_position = $parameters{$key1}{'BP_POSITION'}{1};
	my $rs_number = $parameters{$key1}{'RS_NUMBER'}{1};

	open(FILE2, $file_path) or error ("Error: cannot open file $file_path containing extra information for subpopulation specific local admixture files!");
	while (defined (my $line2 = <FILE2>)) {
	chomp $line2;
	$line2 =~ s/^\s+//g;
	$line2 =~ s/\s+$//g;
	next if ($header eq 'T' && $. == 1);
	my @temp1 = split (/\s+/, $line2);
	$rs_to_bp{$chr}{$temp1[$rs_number-1]} = $temp1[$bp_position-1];
	}
	close FILE2;
	}
	}


	for my $key1 (sort keys %parameters) { # line numbers in parameter file
	if (defined $parameters{$key1}{'LOCALADMIXTUREPERFAMILYFORSUBPOPULATION'}) {
	my $subpop_name = $parameters{$key1}{'LOCALADMIXTUREPERFAMILYFORSUBPOPULATION'}{1};
	my $file_path = $parameters{$key1}{'FILEPATH'}{1};
	$file_path =~ s/^"//g;
	$file_path =~ s/"$//g;
	$file_path =~ s/^'//g;
	$file_path =~ s/'$//g;
	$file_path =~ s/\*/$chr/g;
	my $header = $parameters{$key1}{'HEADER'}{1};
	my $var_id = $parameters{$key1}{'VARIANTIDENTIFIER'}{1};
	my $var_id_col = $parameters{$key1}{'VARIANTIDENTIFIERCOLUMN'}{1};

	my @families;
	my @col_nums;
	for my $key2 (sort keys %{$parameters{$key1}{'FAMILIES'}}) { # consecutive numbers from 1 to the total number of entries
	push @families, $parameters{$key1}{'FAMILIES'}{$key2};
	push @col_nums, $parameters{$key1}{'COLUMNNUMBERS'}{$key2};
	}

	open(FILE1, $file_path) or error ("Error: cannot open file $file_path containing local admixture proportions for reference subpopulation!");
	while (defined (my $line1 = <FILE1>)) {
	chomp $line1;
	$line1 =~ s/^\s+//g;
	$line1 =~ s/\s+$//g;
	next if ($header eq 'T' && $. == 1);
	my @temp1 = split (/\s+/, $line1);

	for (my $el = 0; $el <= $#families; $el++) { 
	if (defined $rs_to_bp{$chr}{$temp1[$var_id_col-1]}) {
	$adms{$chr}{$rs_to_bp{$chr}{$temp1[$var_id_col-1]}}{$subpop_name}{$families[$el]} = $temp1[$col_nums[$el]-1];
	}
	}

	}
	close FILE1;
	}
	}

	return \%adms;
}


sub read_file_admixtures_genome {

	my ($parameters_ref) = $_[0];
	my %parameters = %$parameters_ref;

	my @chrs = (1..22);
	my %adms = ();
	my %rs_to_bp = ();

	for my $key1 (sort keys %parameters) { # line numbers in parameter file
	if (defined $parameters{$key1}{'EXTRAINFORMATIONFORADMIXTUREFILES'}) {
	my $file_path = $parameters{$key1}{'FILEPATH'}{1};
	$file_path =~ s/^"//g;
	$file_path =~ s/"$//g;
	$file_path =~ s/^'//g;
	$file_path =~ s/'$//g;
	my $header = $parameters{$key1}{'HEADER'}{1};
	my $bp_position = $parameters{$key1}{'BP_POSITION'}{1};
	my $rs_number = $parameters{$key1}{'RS_NUMBER'}{1};
	my $chr_number = $parameters{$key1}{'CHR_NUMBERCOLUMN'}{1};

	open(FILE2, $file_path) or error ("Error: cannot open file $file_path containing extra information for subpopulation specific local admixture files!");
	while (defined (my $line2 = <FILE2>)) {
	chomp $line2;
	$line2 =~ s/^\s+//g;
	$line2 =~ s/\s+$//g;
	next if ($header eq 'T' && $. == 1);
	my @temp1 = split (/\s+/, $line2);
	$rs_to_bp{$temp1[$chr_number-1]}{$temp1[$rs_number-1]} = $temp1[$bp_position-1];
	}
	close FILE2;
	}
	}



	for my $chr (@chrs) {
	for my $key1 (sort keys %parameters) { # line numbers in parameter file
	if (defined $parameters{$key1}{'LOCALADMIXTUREPERFAMILYFORSUBPOPULATION'}) {
	my $subpop_name = $parameters{$key1}{'LOCALADMIXTUREPERFAMILYFORSUBPOPULATION'}{1};
	my $file_path = $parameters{$key1}{'FILEPATH'}{1};
	$file_path =~ s/^"//g;
	$file_path =~ s/"$//g;
	$file_path =~ s/^'//g;
	$file_path =~ s/'$//g;
	$file_path =~ s/\*/$chr/g;
	my $header = $parameters{$key1}{'HEADER'}{1};
	my $var_id = $parameters{$key1}{'VARIANTIDENTIFIER'}{1};
	my $var_id_col = $parameters{$key1}{'VARIANTIDENTIFIERCOLUMN'}{1};

	my @families;
	my @col_nums;
	for my $key2 (sort keys %{$parameters{$key1}{'FAMILIES'}}) { # consecutive numbers from 1 to the total number of entries
	push @families, $parameters{$key1}{'FAMILIES'}{$key2};
	push @col_nums, $parameters{$key1}{'COLUMNNUMBERS'}{$key2};
	}

	open(FILE1, $file_path) or error ("Error: cannot open file $file_path containing local admixture proportions for reference subpopulation!");
	while (defined (my $line1 = <FILE1>)) {
	chomp $line1;
	$line1 =~ s/^\s+//g;
	$line1 =~ s/\s+$//g;
	next if ($header eq 'T' && $. == 1);
	my @temp1 = split (/\s+/, $line1);

	for (my $el = 0; $el <= $#families; $el++) { 
	if (defined $rs_to_bp{$chr}{$temp1[$var_id_col-1]}) {
	$adms{$chr}{$rs_to_bp{$chr}{$temp1[$var_id_col-1]}}{$subpop_name}{$families[$el]} = $temp1[$col_nums[$el]-1];
	}
	}

	}
	close FILE1;
	}
	}
	}

	return \%adms;
}


sub read_geno_file_info {

	my ($parameters_ref, $chr_ref, $family_ref) = @_;
	my %parameters = %$parameters_ref;
	my $chr = $$chr_ref;
	my $family = $$family_ref;

	my %marker_cm_pos = ();
	my %marker_info = ();
	my %genotypes = ();
	my $geno_line_1;

	my $file_path;
	for my $key1 (sort keys %parameters) { # line numbers in parameter file
	for my $key2 (sort keys %{$parameters{$key1}}) { # options from specific line of parameter file
	if ($key2 eq 'GENOFILEFORGL_AUTO') {
	$file_path = $parameters{$key1}{'FILEPATH'}{1};
	}
	}
	}

	$file_path =~ s/^"//g;
	$file_path =~ s/"$//g;
	$file_path =~ s/^'//g;
	$file_path =~ s/'$//g;
	$file_path =~ s/\*/$chr/g;
	$file_path =~ s/\+/$family/g;

	open(FILE1, $file_path) or error ("Error: cannot open geno file $file_path!");
	while (defined (my $line1 = <FILE1>)) {
	chomp $line1;
	$line1 =~ s/^\s+//g;
	$line1 =~ s/\s+$//g;

	next if (!$line1);
	$geno_line_1 = $line1 if ($line1 =~ m/map/);
	my @temp1 = split (/\D+\s+/, $line1) if ($. == 1);
	my @temp2 = split (/\s+/, $temp1[1]) if ($. == 1);
	for my $el (@temp2) {
	$marker_cm_pos{$el} = 1;
	}

	if ($line1 !~ m/marker/ && $. != 1) {
	my @temp3 = split (/\s+/, $line1);
	$genotypes{$.} = [@temp3];
	}


	}
	close FILE1;


	my $file_path_2;
	for my $key1 (sort keys %parameters) {
	for my $key2 (sort keys %{$parameters{$key1}}) {
	if ($key2 eq 'INFORMATIONFORGENOFILE') {
	$file_path_2 = $parameters{$key1}{'FILEPATH'}{1};
	error ("Error: location of file with extra information for geno file is not provided!") if (!defined $file_path_2);
	$file_path_2 =~ s/^"//g;
	$file_path_2 =~ s/"$//g;
	$file_path_2 =~ s/^'//g;
	$file_path_2 =~ s/'$//g;
	$file_path_2 =~ s/\*/$chr/g;
	$file_path_2 =~ s/\+/$family/g;

	my $header = $parameters{$key1}{'HEADER'}{1};
	my $bp_position = $parameters{$key1}{'BP_POSITION'}{1};
	my $cm_position = $parameters{$key1}{'CM_POSITION'}{1};
	my $rs_number = $parameters{$key1}{'RS_NUMBER'}{1};


	open(FILE2, $file_path_2) or error ("Error: cannot open file $file_path containing additional information for geno file!");
	while (defined (my $line2 = <FILE2>)) {
	chomp $line2;
	$line2 =~ s/^\s+//g;
	$line2 =~ s/\s+$//g;
	next if ($header eq 'T' && $. == 1);
	my @temp1 = split (/\s+/, $line2);
	$marker_info{$temp1[$bp_position-1]}{'CM_POSITION'} = $temp1[$cm_position-1];
	$marker_info{$temp1[$bp_position-1]}{'RS_NUMBER'} = $temp1[$rs_number-1];
	}
	close FILE2;

	}
	}
	} # 2nd major loop end


	my @geno_bp;
	for my $key1 (sort {$a <=> $b} keys %marker_info) {
	if (!defined $marker_cm_pos{$marker_info{$key1}{'CM_POSITION'}} ) {
#	print "problem!!!!!!!!!!!!\n";
	} else {
	push @geno_bp, $key1;
	}
	}

	return \%marker_cm_pos, \@geno_bp, \%genotypes, \$geno_line_1;
}


sub read_file_sample_allele_frqs {

	my ($parameters_ref, $chr_ref) = @_;
	my $chr = $$chr_ref;
	my %parameters = %$parameters_ref;

	my %sample_frqs = ();

	for my $key1 (sort keys %parameters) { # line numbers in parameter file
	if (defined $parameters{$key1}{'YOURSAMPLEALLELEFREQUENCIES'}) {
	my $file_path = $parameters{$key1}{'FILEPATH'}{1};
	$file_path =~ s/^"//g;
	$file_path =~ s/"$//g;
	$file_path =~ s/^'//g;
	$file_path =~ s/'$//g;
	$file_path =~ s/\*/$chr/g;
	my $header = $parameters{$key1}{'HEADER'}{1};
	my $bp_position = $parameters{$key1}{'BP_POSITION'}{1};
	my $al1 = $parameters{$key1}{'ALELLE1'}{1};
	my $al1_frq = $parameters{$key1}{'FREQUENCYOFALLELE1'}{1};
	my $al2 = $parameters{$key1}{'ALELLE2'}{1};
	my $al2_frq = $parameters{$key1}{'FREQUENCYOFALLELE2'}{1};
	open(FILE1, $file_path) or error ("Error: cannot open file $file_path containing your sample specific allele frequencies!");
	while (defined (my $line1 = <FILE1>)) {
	chomp $line1;
	$line1 =~ s/^\s+//g;
	$line1 =~ s/\s+$//g;
	next if ($header eq 'T' && $. == 1);
	my @temp1 = split (/\s+/, $line1);
	$sample_frqs{$chr}{$temp1[$bp_position-1]}{'al1'} = $temp1[$al1-1];
	$sample_frqs{$chr}{$temp1[$bp_position-1]}{'al1_frq'} = $temp1[$al1_frq-1];
	$sample_frqs{$chr}{$temp1[$bp_position-1]}{'al2'} = $temp1[$al2-1];
	$sample_frqs{$chr}{$temp1[$bp_position-1]}{'al2_frq'} = $temp1[$al2_frq-1];
	}
	close FILE1;

	}
	}

return \%sample_frqs;
}


sub read_bp_modify_frq {

	my ($parameters_ref, $chr_ref) = @_;
	my $chr = $$chr_ref;
	my %parameters = %$parameters_ref;

	my @bp_pos = ();
	
	for my $key1 (sort keys %parameters) { # line numbers in parameter file
	if (defined $parameters{$key1}{'MODIFYALLELEFREQUENCIES'}) {
	my $file_path = $parameters{$key1}{'FILEPATH'}{1};
	$file_path =~ s/^"//g;
	$file_path =~ s/"$//g;
	$file_path =~ s/^'//g;
	$file_path =~ s/'$//g;
	$file_path =~ s/\*/$chr/g;
	my $header = $parameters{$key1}{'HEADER'}{1};
	my $bp_position = $parameters{$key1}{'BP_POSITION'}{1};

	open(FILE1, $file_path) or error ("Error: cannot open file $file_path containing base pair positions for which allele frequencies were requested to be modified!");
	while (defined (my $line1 = <FILE1>)) {
	chomp $line1;
	$line1 =~ s/^\s+//g;
	$line1 =~ s/\s+$//g;
	next if ($header eq 'T' && $. == 1);
	my @temp1 = split (/\s+/, $line1);
	push @bp_pos, $temp1[$bp_position-1];
	
	}
	close FILE1;
	}
	}

return \@bp_pos;
}


sub make_dir {
	my ($parameters_ref, $chr_ref, $family_ref, $ancestry_model_ref, $comp_type_ref) = @_;
# $comp_type = 0 if frequencies for geno files are needed; 1 = if frequencies not for geno files are needed (FBGW, FBCW, Local); 2 = if admixture proportions for Local model are needed; 3 = if admixture proportions for Global, FBGW, or FBCW model are needed; 4 = directory for error log file; 5 = if frequencies not for geno files are needed (Global); 6 = printout of local admixture proportions for each variant in geno file
	my %parameters = %$parameters_ref;
	my $chr = $$chr_ref if ($chr_ref != 0);
	my $family = $$family_ref if ($family_ref != 0);
	my $ancestry_model = $$ancestry_model_ref if ($ancestry_model_ref != 0);
	my $comp_type = $$comp_type_ref;

	my $dir_out;
	for my $key1 (sort keys %parameters) { # line numbers in parameter file
	if (defined $parameters{$key1}{'OUTPUTDIRECTORY'}) {
	$dir_out = $parameters{$key1}{'OUTPUTDIRECTORY'}{1};
	$dir_out =~ s/^"//g;
	$dir_out =~ s/"$//g;
	$dir_out =~ s/^'//g;
	$dir_out =~ s/'$//g;
	}
	}

	if (!defined $dir_out) {error_2 ("\n	Error: Output directory is not specified!\n	Log file was not created!\n")};

	my $dir0;
	my $dir1;
	my $dir2;
	my $dir3;
	my $dir4;
	my $dir5;
	my $dir6;

	if ($comp_type == 0) {
	$dir0 = "$dir_out/gl_auto/$ancestry_model/$family/chr$chr";
	if (! -e $dir0) {
	`mkdir -p $dir0`;
	}
	return \$dir0;
	}

	elsif ($comp_type == 1) {
	$dir1 = "$dir_out/allele_frequencies/$ancestry_model/$family/chr$chr";
	if (! -e $dir1) {
	`mkdir -p $dir1`;
	}
	return \$dir1;
	}

	elsif ($comp_type == 2) {
	$dir2 = "$dir_out/admixture_proportions/$ancestry_model/$family/chr$chr";
	if (! -e $dir2) {
	`mkdir -p $dir2`;
	}
	return \$dir2;
	}

	elsif ($comp_type == 3) {
	$dir3 = "$dir_out/admixture_proportions/$ancestry_model";
	if (! -e $dir3) {
	`mkdir -p $dir3`;
	}
	return \$dir3;
	}

	elsif ($comp_type == 4) {
	$dir4 = "$dir_out";
	if (! -e $dir4) {
	`mkdir -p $dir4`;
	}
	return \$dir4;
	}

	elsif ($comp_type == 5) {
	$dir5 = "$dir_out/allele_frequencies/$ancestry_model/chr$chr";
	if (! -e $dir5) {
	`mkdir -p $dir5`;
	}
	return \$dir5;
	}

	elsif ($comp_type == 6) {
	$dir6 = "$dir_out/gl_auto_local_adm/$family/chr$chr";
	if (! -e $dir6) {
	`mkdir -p $dir6`;
	}
	return \$dir6;
	}

}


sub print_geno_file {

	my ($geno_genotypes_ref, $geno_line_1_ref, $geno_frqs_ref, $parameters_ref, $chr_ref, $family_ref, $ancestry_m) = @_;

	my %geno_genotypes = %$geno_genotypes_ref;
	my $geno_line_1 = $$geno_line_1_ref;
	my %geno_frqs = %$geno_frqs_ref;
	my %parameters = %$parameters_ref;
	my $chr = $$chr_ref;
	my $family = $$family_ref;

	my $ancestry_model; # 1 = Global; 2 = FBGW; 3 = FBCW; 4 = Local
	if ($ancestry_m == 1) {
	$ancestry_model = 'Global';
	}
	elsif ($ancestry_m == 2) {
	$ancestry_model = 'FBGW';
	}
	elsif ($ancestry_m == 3) {
	$ancestry_model = 'FBCW';
	}
	elsif ($ancestry_m == 4) {
	$ancestry_model = 'Local';
	}
	my $ancestry_model_ref = \$ancestry_model;

	my $comp_type = 0;
	my $comp_type_ref = \$comp_type;
	my $dir_ref = make_dir ($parameters_ref, $chr_ref, $family_ref, $ancestry_model_ref, $comp_type_ref);
	my $dir = $$dir_ref;

	open (FILEOUT1, ">", "$dir/chr$chr.geno");
	print FILEOUT1 "$geno_line_1\n\n";


	my @geno_frqs_out;
	for my $key1 (sort {$a <=> $b} keys %{$geno_frqs{$chr}}) { # bp
	push @geno_frqs_out, "$geno_frqs{$chr}{$key1}{1}[1]\n";
	}
	my $number_m = scalar @geno_frqs_out;

	for (my $el = 0; $el <= $#geno_frqs_out; $el++) {
	my $temp1 = $el + 1;
	my $temp2 = sprintf ("%.6f", $geno_frqs_out[$el]);
	my $temp3 = 1 - $temp2;
	$temp3 = sprintf ("%.6f", $temp3);
	print FILEOUT1 "set markers $temp1 allele freq $temp2 $temp3\n";
	}
	print FILEOUT1 "\n";
	print FILEOUT1 "set markers $number_m data\n\n";

	for my $key1 (sort {$a <=> $b} keys %geno_genotypes) {
	print FILEOUT1 "@{$geno_genotypes{$key1}}\n";
	}

close FILEOUT1;

}


sub print_geno_file_frqs {

	my ($marker_cm_pos_ref, $geno_frqs_ref, $parameters_ref, $chr_ref, $family_ref, $ancestry_m) = @_;
	my %marker_cm_pos = %$marker_cm_pos_ref;
	my %geno_frqs = %$geno_frqs_ref;
	my %parameters = %$parameters_ref;
	my $chr = $$chr_ref;
	my $family = $$family_ref;

	my $ancestry_model; # 1 = Global; 2 = FBGW; 3 = FBCW; 4 = Local
	if ($ancestry_m == 1) {
	$ancestry_model = 'Global';
	}
	elsif ($ancestry_m == 2) {
	$ancestry_model = 'FBGW';
	}
	elsif ($ancestry_m == 3) {
	$ancestry_model = 'FBCW';
	}
	elsif ($ancestry_m == 4) {
	$ancestry_model = 'Local';
	}
	my $ancestry_model_ref = \$ancestry_model;

	my $comp_type = 0;
	my $comp_type_ref = \$comp_type;
	my $dir_ref = make_dir ($parameters_ref, $chr_ref, $family_ref, $ancestry_model_ref, $comp_type_ref);
	my $dir = $$dir_ref;

	open (FILEOUT1, ">", "$dir/chr$chr.frq");
	print FILEOUT1 "bp_position cM_position al1 al2 al1_frq al2_frq\n";

	my @cm_pos;
	for my $key1 (sort {$a <=> $b} keys %marker_cm_pos) { # cM positions
	push @cm_pos, $key1;
	}

	my $count = 0;
	for my $key1 (sort {$a <=> $b} keys %{$geno_frqs{$chr}}) { # bp position
	print FILEOUT1 " $key1 $cm_pos[$count] $geno_frqs{$chr}{$key1}{1}[0] $geno_frqs{$chr}{$key1}{2}[0] $geno_frqs{$chr}{$key1}{1}[1] $geno_frqs{$chr}{$key1}{2}[1]\n";
	$count++;
	}

}


sub print_local_adm_geno_file {

	my ($marker_cm_pos_ref, $local_adms_geno_ref, $parameters_ref, $chr_ref, $family_ref) = @_;
	my %marker_cm_pos = %$marker_cm_pos_ref;
	my %local_adms_geno = %$local_adms_geno_ref;
	my %parameters = %$parameters_ref;
	my $chr = $$chr_ref;
	my $family = $$family_ref;

	my $ancestry_model = 'Local';
	my $ancestry_model_ref = \$ancestry_model;
	my $comp_type = 6;
	my $comp_type_ref = \$comp_type;
	my $dir_ref = make_dir ($parameters_ref, $chr_ref, $family_ref, $ancestry_model_ref, $comp_type_ref);
	my $dir = $$dir_ref;

	open (FILEOUT1, ">", "$dir/chr$chr.local.adm.txt");
	print FILEOUT1 "bp_position cM_position";

	my @cm_pos;
	for my $key1 (sort {$a <=> $b} keys %marker_cm_pos) {
	push @cm_pos, $key1;
	}

	for my $key1 (sort {$a <=> $b} keys %{$local_adms_geno{$chr}}) {
	for my $key2 (sort {$a cmp $b} keys %{$local_adms_geno{$chr}{$key1}}) {
	print FILEOUT1 " local_adm_$key2";  
	}
	last;
	}
	print FILEOUT1 "\n";

	my $count = 0;
	for my $key1 (sort {$a <=> $b} keys %{$local_adms_geno{$chr}}) { # bp position
	print FILEOUT1 " $key1 $cm_pos[$count]";
	$count++;
	for my $key2 (sort {$a cmp $b} keys %{$local_adms_geno{$chr}{$key1}}) { # subpopulation
	if ($local_adms_geno{$chr}{$key1}{$key2} == 0 || $local_adms_geno{$chr}{$key1}{$key2} == 1) {
	my $temp0 = sprintf ("%.0f", $local_adms_geno{$chr}{$key1}{$key2});
	print FILEOUT1 " $temp0";
	} else {
	my $temp1 = sprintf ("%.6f", $local_adms_geno{$chr}{$key1}{$key2});
	print FILEOUT1 " $temp1";
	}
	}
		print FILEOUT1 "\n";
	}

}


sub print_local_adm_file {

	my ($local_adms_ref, $subpop_local_ref, $parameters_ref, $chr_ref, $family_ref) = @_;
	my %local_adms = %$local_adms_ref;
	my %subpop_local = %$subpop_local_ref;
	my %parameters = %$parameters_ref;
	my $chr = $$chr_ref;
	my $family = $$family_ref;

	my $ancestry_model = 'Local';
	my $ancestry_model_ref = \$ancestry_model;
	my $comp_type = 2;
	my $comp_type_ref = \$comp_type;
	my $dir_ref = make_dir ($parameters_ref, $chr_ref, $family_ref, $ancestry_model_ref, $comp_type_ref);
	my $dir = $$dir_ref;

	open (FILEOUT1, ">", "$dir/chr$chr.$family.txt");
	print FILEOUT1 "bp_position";

	for my $key (sort { $subpop_local{$a} <=> $subpop_local{$b} } keys %subpop_local) {
	print FILEOUT1 " local_adm_$key";  
	}
	print FILEOUT1 "\n";


	for my $key ( sort {$a <=> $b} keys %local_adms) {
	my @temp1 = split (/\s+/, "@{$local_adms{$key}}");
	my @temp2;
	for my $el (@temp1) {
	if ($el == 0 || $el == 1) {
	push @temp2, $el;
	} else {
	push @temp2, sprintf ("%.6f", $el);
	}
	}
	print FILEOUT1 "$key @temp2\n";
	}
	close FILEOUT1;

}


sub remove_old_fbcw_adm_file {

	my ($parameters_ref, $chr_ref, $family_ref) = @_;
	my %parameters = %$parameters_ref;

	my $ancestry_model = 'FBCW';
	my $ancestry_model_ref = \$ancestry_model;
	my $comp_type = 3;
	my $comp_type_ref = \$comp_type;
	my $dir_ref = make_dir ($parameters_ref, $chr_ref, $family_ref, $ancestry_model_ref, $comp_type_ref);
	my $dir = $$dir_ref;
	open (FILEOUT1, ">", "$dir/fbcw.admixture.txt");
	close FILEOUT1;

}
 

sub print_fbcw_adm_file {

	my ($fbcw_adm_ref, $parameters_ref, $chr_ref, $family_ref) = @_;
	my %fbcw_adm = %$fbcw_adm_ref;
	my %parameters = %$parameters_ref;

	my $ancestry_model = 'FBCW';
	my $ancestry_model_ref = \$ancestry_model;
	my $comp_type = 3;
	my $comp_type_ref = \$comp_type;
	my $dir_ref = make_dir ($parameters_ref, $chr_ref, $family_ref, $ancestry_model_ref, $comp_type_ref);
	my $dir = $$dir_ref;

	my @subpop;
	for my $key1 (sort {$a <=> $b} keys %fbcw_adm) { # chr
	for my $key2 (sort {$a cmp $b} keys %{$fbcw_adm{$key1}}) {# family
	for my $key3 (sort {$a cmp $b} keys %{$fbcw_adm{$key1}{$key2}}) {# subpopulation
	push @subpop, $key3;
	}
	last;
	}
	last;
	}

	open (FILEOUT1, ">", "$dir/fbcw.admixture.txt") if ($chr_ref == 0);
	open (FILEOUT1, ">>", "$dir/fbcw.admixture.txt") if ($chr_ref != 0);
	print FILEOUT1 "chr family";
	for my $el (@subpop) {
	print FILEOUT1 " FBCW_adm_$el";
	}
	print FILEOUT1 "\n";

	for my $key1 (sort {$a <=> $b} keys %fbcw_adm) { # chr
	for my $key2 (sort {$a cmp $b} keys %{$fbcw_adm{$key1}}) {# family
	print FILEOUT1 "$key1 $key2";
	for my $key3 (sort {$a cmp $b} keys %{$fbcw_adm{$key1}{$key2}}) {# subpopulation
	if ($fbcw_adm{$key1}{$key2}{$key3} == 0 || $fbcw_adm{$key1}{$key2}{$key3} == 1) {
	print FILEOUT1 " $fbcw_adm{$key1}{$key2}{$key3}"; 
	} else {
	my $temp1 = sprintf ("%.6f", $fbcw_adm{$key1}{$key2}{$key3});
	print FILEOUT1 " $temp1";
	}
	}
	print FILEOUT1 "\n";
	}
	}
	close FILEOUT1;

}


sub print_fbgw_adm_file {

	my ($fbgw_adm_ref, $parameters_ref) = @_;
	my %fbgw_adm = %$fbgw_adm_ref;
	my $ancestry_model = 'FBGW';
	my $ancestry_model_ref = \$ancestry_model;
	my $comp_type = 3;
	my $comp_type_ref = \$comp_type;
	my $dir_ref = make_dir ($parameters_ref, 0, 0, $ancestry_model_ref, $comp_type_ref);
	my $dir = $$dir_ref;

	my @subpop;
	for my $key1 ( sort {$a cmp $b} keys %fbgw_adm) {
	for my $key2 ( sort {$a cmp $b} keys %{ $fbgw_adm{$key1}} ) {
	push @subpop, $key2;
	}
	last;
	}

	open (FILEOUT1, ">", "$dir/fbgw.admixture.txt");
	print FILEOUT1 "family";
	for my $el (@subpop) {
	print FILEOUT1 " FBGW_adm_$el";
	}
	print FILEOUT1 "\n";


	for my $key1 ( sort {$a cmp $b} keys %fbgw_adm) {
	print FILEOUT1 "$key1";
	for my $key2 ( sort {$a cmp $b} keys %{ $fbgw_adm{$key1}} ) {

	if ($fbgw_adm{$key1}{$key2} == 0 || $fbgw_adm{$key1}{$key2} == 1) {
	print FILEOUT1 " $fbgw_adm{$key1}{$key2}";
	} else {
	my $temp1 = sprintf ("%.6f", $fbgw_adm{$key1}{$key2});
	print FILEOUT1 " $temp1";
	}
	}
	print FILEOUT1 "\n";
	}
	close FILEOUT1;

}


sub print_global_adm_file {

	my ($global_ref, $parameters_ref) = @_;
	my %global = %$global_ref;
	my %parameters_ref = %$parameters_ref;

	my $ancestry_model = 'Global';
	my $ancestry_model_ref = \$ancestry_model;
	my $comp_type = 3;
	my $comp_type_ref = \$comp_type;
	my $dir_ref = make_dir ($parameters_ref, 0, 0, $ancestry_model_ref, $comp_type_ref);
	my $dir = $$dir_ref;

	open (FILEOUT1, ">", "$dir/global.admixture.txt");
	print FILEOUT1 "subpopulation global_adm\n";
	for my $key1 ( sort {$a cmp $b} keys %global) {
	if ($global{$key1} == 0 || $global{$key1} == 1) {
	print FILEOUT1 "$key1 $global{$key1}\n";
	} else {
	my $temp1 = sprintf ("%.6f", $global{$key1});
	print FILEOUT1 "$key1 $temp1\n";
	}
	}
	close FILEOUT1;

}


sub print_frqs {

	my ($parameters_ref, $frqs_ref, $chr_ref, $family_ref, $ancestry_m) = @_;
	my %parameters = %$parameters_ref;
	my %frqs = %$frqs_ref;
	my $chr = $$chr_ref;

	my $ancestry_model; # 1 = Global; 2 = FBGW; 3 = FBCW; 4 = Local
	if ($ancestry_m == 1) {
	$ancestry_model = 'Global';
	}
	elsif ($ancestry_m == 2) {
	$ancestry_model = 'FBGW';
	}
	elsif ($ancestry_m == 3) {
	$ancestry_model = 'FBCW';
	}
	elsif ($ancestry_m == 4) {
	$ancestry_model = 'Local';
	}
	my $ancestry_model_ref = \$ancestry_model;

	my $comp_type;
	$comp_type = 1;
	$comp_type = 5 if ($ancestry_m == 1);
	my $comp_type_ref = \$comp_type;

	my $dir_ref = make_dir ($parameters_ref, $chr_ref, $family_ref, $ancestry_model_ref, $comp_type_ref);
	my $dir = $$dir_ref;

	open (FILEOUT1, ">", "$dir/global.frq.txt") if ($ancestry_m == 1);
	print FILEOUT1 "bp_position al1 al2 al1_frq al2_frq\n" if ($ancestry_m == 1);

	open (FILEOUT1, ">", "$dir/fbgw.frq.txt") if ($ancestry_m == 2);
	print FILEOUT1 "bp_position al1 al2 al1_frq al2_frq\n" if ($ancestry_m == 2);

	open (FILEOUT1, ">", "$dir/fbcw.frq.txt") if ($ancestry_m == 3);
	print FILEOUT1 "bp_position al1 al2 al1_frq al2_frq\n" if ($ancestry_m == 3);

	open (FILEOUT1, ">", "$dir/local.frq.txt") if ($ancestry_m == 4);
	print FILEOUT1 "bp_position al1 al2 al1_frq al2_frq\n" if ($ancestry_m == 4);

	for my $key1 (sort {$a <=> $b} keys %{ $frqs{$chr}}) { # bp position
	print FILEOUT1 "$key1 $frqs{$chr}{$key1}{1}[0] $frqs{$chr}{$key1}{2}[0] $frqs{$chr}{$key1}{1}[1] $frqs{$chr}{$key1}{2}[1]\n";
	}
	close FILEOUT1;

}


sub adms_fbcw { # compute FBCW admixture proportions only for allele frequency computations

	my $adm_ref = $_[0];
	my %adm = %$adm_ref;
	my %fbcw = ();
	my %fbcw_out = ();

	for my $key1 (sort keys %adm) { # chr
	for my $key2 (sort keys %{ $adm{$key1} }) { # bp position
	for my $key3 (sort keys %{ $adm{$key1}{$key2} }) { # subpopulation
	for my $key4 (sort keys %{ $adm{$key1}{$key2}{$key3} }) { # family
	push @{$fbcw{$key1}{$key4}{$key3}}, $adm{$key1}{$key2}{$key3}{$key4};
	}
	}
	}
	}


	for my $key1 (sort keys %fbcw) {
	for my $key2 (sort keys %{ $fbcw{$key1} } ) {
	for my $key3 (sort keys %{ $fbcw{$key1}{$key2} } ) { # $key1 = chr, $key2 = family, $key3 = population
	my @temp1 = @{$fbcw{$key1}{$key2}{$key3}};
	my $sum;
	foreach (@temp1) {
	$sum += $_;
	}
	my $result = $sum/ scalar @temp1;
	$fbcw_out{$key1}{$key2}{$key3} = $result;
	}
	}
	}

return \%fbcw_out;
}


sub adm_global_supplied {

	my $parameters_ref = $_[0];
	my %parameters = %$parameters_ref;

	my %global_adm = ();
	for my $key1 (sort keys %parameters) {
	for my $key2 (sort keys %{$parameters{$key1}}) {
	if ($key2 eq 'GLOBALANCESTRYMODELPROPORTIONS') {
	for my $key3 (sort keys %{$parameters{$key1}}) {
	$global_adm{$key3} = $parameters{$key1}{$key3}{1} if ($key3 ne 'GLOBALANCESTRYMODELPROPORTIONS');
	}
	}
	}
	}

	return \%global_adm;
}


sub adms_global_fbgw { # computes Global and FBGW admixture proportions for print out into a file and for allele frequency computations

	my ($average_type_ref, $adm_ref, $global_adm_comp_ref) = @_;
	my $average_type = $$average_type_ref; # 1 = weighted, 2 = normal
	my %adm = %$adm_ref;
	my $global_adm_comp = $$global_adm_comp_ref; # 0 = use specified Global admixture; 1 = compute Global admixture from the data

	my %fbgw = (); # weighted
	my %fbgw_out = (); # weighted
	my %fbcw = ();
	my %fbcw_out = ();
	my %fbgw_n = (); # normal
	my %fbgw_n_out = (); # normal
	my %global = ();
	my %global_out = ();
	my %global_n = ();
	my %global_n_out = ();


	for my $key1 (sort keys %adm) { # chr
	for my $key2 (sort keys %{ $adm{$key1} }) { # bp position
	for my $key3 (sort keys %{ $adm{$key1}{$key2} }) { # subpopulation
	for my $key4 (sort keys %{ $adm{$key1}{$key2}{$key3} }) { # family
	push @{$fbgw{$key4}{$key3}}, $adm{$key1}{$key2}{$key3}{$key4} if ($average_type == 1);
	push @{$fbcw{$key1}{$key4}{$key3}}, $adm{$key1}{$key2}{$key3}{$key4} if ($average_type == 2);
	}
	}
	}
	}


	if ($average_type == 2) {

	for my $key1 (sort keys %fbcw) { # chr
	for my $key2 (sort keys %{ $fbcw{$key1} } ) { # family
	for my $key3 (sort keys %{ $fbcw{$key1}{$key2} } ) { # subpopulation
	my @temp1 = @{$fbcw{$key1}{$key2}{$key3}};
	my $sum;
	foreach (@temp1) {
	$sum += $_;
	}
	my $result = $sum/ scalar @temp1;
	$fbcw_out{$key1}{$key2}{$key3} = $result;
	}
	}
	}

	for my $key1 (sort keys %fbcw_out) { # chr
	for my $key2 (sort keys %{ $fbcw_out{$key1} } ) { # family
	for my $key3 (sort keys %{ $fbcw_out{$key1}{$key2} } ) { # subpopulation
	push @{$fbgw_n{$key2}{$key3}}, $fbcw_out{$key1}{$key2}{$key3};
	}
	}
	}


	for my $key1 (sort keys %fbgw_n) { # family
	for my $key2 (sort keys %{ $fbgw_n{$key1} } ) { # subpopulation
	my @temp1 = @{$fbgw_n{$key1}{$key2}};
	my $sum;
	foreach (@temp1) {
	$sum += $_;
	}
	my $result = $sum/ scalar @temp1;
	$fbgw_n_out{$key1}{$key2} = $result;
	}
	}


	if ($global_adm_comp == 1 ) {
	for my $key1 (sort keys %fbgw_n_out) { # family
	for my $key2 (sort keys %{ $fbgw_n_out{$key1} }) { # subpopulation
	push @{$global_n{$key2}}, $fbgw_n_out{$key1}{$key2};
	}
	}


	for my $key1 (sort keys %global_n) { # subpopulation
	my @temp1 = @{$global_n{$key1}};

	my $sum;
	foreach (@temp1) {
	$sum += $_;
	}
	my $result = $sum/ scalar @temp1;
	$global_n_out{$key1} = $result;
	}
	}

	}


	if ($average_type == 1) {
	for my $key1 (sort keys %fbgw) { # family
	for my $key2 (sort keys %{ $fbgw{$key1} }) { # subpopulation
	my @temp1 = @{$fbgw{$key1}{$key2}};

	my $sum;
	foreach (@temp1) {
	$sum += $_;
	}
	my $result = $sum/ scalar @temp1;
	$fbgw_out{$key1}{$key2} = $result;
	}
	}


	for my $key1 (sort keys %fbgw) { # family
	for my $key2 (sort keys %{ $fbgw{$key1} }) { # subpopulation
	push @{$global{$key2}}, @{$fbgw{$key1}{$key2}};
	}
	}


	if ($global_adm_comp == 1) {
	for my $key1 (sort keys %global) { # subpopulation
	my @temp1 = @{$global{$key1}};

	my $sum;
	foreach (@temp1) {
	$sum += $_;
	}
	my $result = $sum/ scalar @temp1;
	$global_out{$key1} = $result;
	}
	}

	}

	return \(%fbgw_out, %global_out) if ($average_type == 1 && $global_adm_comp == 1);
	return \(%fbgw_n_out, %global_n_out) if ($average_type == 2 && $global_adm_comp == 1);

	return \%fbgw_out if ($average_type == 1 && $global_adm_comp == 0);
	return \%fbgw_n_out if ($average_type == 2 && $global_adm_comp == 0);

}


sub adms_fbcw_b {

	my ($adm_ref) = @_;
	my %adm = %$adm_ref;

	my %fbcw = ();
	my %fbcw_out = ();

	for my $key1 (sort keys %adm) { # chr
	for my $key2 (sort keys %{ $adm{$key1} }) { # bp position
	for my $key3 (sort keys %{ $adm{$key1}{$key2} }) { # subpopulation
	for my $key4 (sort keys %{ $adm{$key1}{$key2}{$key3} }) { # family
	push @{$fbcw{$key1}{$key4}{$key3}}, $adm{$key1}{$key2}{$key3}{$key4};
	}
	}
	}
	}


	for my $key1 (sort keys %fbcw) { # chr
	for my $key2 (sort keys %{ $fbcw{$key1} } ) { # family
	for my $key3 (sort keys %{ $fbcw{$key1}{$key2} } ) { # subpopulation
	my @temp1 = @{$fbcw{$key1}{$key2}{$key3}};
	my $sum;
	foreach (@temp1) {
	$sum += $_;
	}
	my $result = $sum/ scalar @temp1;
	$fbcw_out{$key1}{$key2}{$key3} = $result;
	}
	}
	}
	return \%fbcw_out;
}


sub frq_template {

	my ($parameters_ref, $sample_frqs_ref, $pop_frqs_ref, $bp_pos_for_frq_ref, $chr_num_ref) = @_;
	my %parameters = %$parameters_ref;
	my %sample_frqs = %$sample_frqs_ref;
	my %pop_frqs = %$pop_frqs_ref;
	my @bp_pos_for_frq = @$bp_pos_for_frq_ref;
	my $chr_num = $$chr_num_ref;


 	my %subpop = ();
	for my $key1 (sort keys %parameters) {
	if (defined $parameters{$key1}{'SUBPOPULATIONNAME'}) {
	$subpop{$parameters{$key1}{'SUBPOPULATIONNAME'}{1}} = 1;
	}
	}
# maybe change to sort by value
	my @subpop_frq_template;
	for my $sbp (sort keys %subpop) {
	push @subpop_frq_template, $sbp;
	}

	my %frq_template = ();

	for my $el1 (sort @bp_pos_for_frq) {
	if (defined $sample_frqs{$chr_num}{$el1}) {
	for my $sbp (sort keys %subpop) {
	if (!defined $pop_frqs{$chr_num}{$el1}{$sbp} && !exists $frq_template{$chr_num}{$el1}{1}) {
	if ($sample_frqs{$chr_num}{$el1}{'al1_frq'} >= 0.5) {
	push @{$frq_template{$chr_num}{$el1}{1}}, $sample_frqs{$chr_num}{$el1}{'al1'}, 0.99;
	$frq_template{$chr_num}{$el1}{2} = $sample_frqs{$chr_num}{$el1}{'al2'};
	}
	elsif ($sample_frqs{$chr_num}{$el1}{'al2_frq'} > 0.5) {
	push @{$frq_template{$chr_num}{$el1}{1}}, $sample_frqs{$chr_num}{$el1}{'al2'}, 0.99;
	$frq_template{$chr_num}{$el1}{2} = $sample_frqs{$chr_num}{$el1}{'al1'};
	}
	}


	elsif (defined $pop_frqs{$chr_num}{$el1}{$sbp} && !exists $frq_template{$chr_num}{$el1}{1}) {
	push @{$frq_template{$chr_num}{$el1}{1}}, $pop_frqs{$chr_num}{$el1}{$sbp}{'REF_ALLELE'};
	push @{$frq_template{$chr_num}{$el1}{1}}, $pop_frqs{$chr_num}{$el1}{$sbp}{'REF_ALLELE_FRQ'};
	$frq_template{$chr_num}{$el1}{2} = $pop_frqs{$chr_num}{$el1}{$sbp}{'ALT_ALLELE'};
	}


	elsif (defined $pop_frqs{$chr_num}{$el1}{$sbp} && exists $frq_template{$chr_num}{$el1}{1}) {
	if ($pop_frqs{$chr_num}{$el1}{$sbp}{'REF_ALLELE'} eq @{$frq_template{$chr_num}{$el1}{1}}[0]) {
	push @{$frq_template{$chr_num}{$el1}{1}}, $pop_frqs{$chr_num}{$el1}{$sbp}{'REF_ALLELE'};
	push @{$frq_template{$chr_num}{$el1}{1}}, $pop_frqs{$chr_num}{$el1}{$sbp}{'REF_ALLELE_FRQ'};
	}
	elsif ($pop_frqs{$chr_num}{$el1}{$sbp}{'ALT_ALLELE'} eq @{$frq_template{$chr_num}{$el1}{1}}[0]) {
	push @{$frq_template{$chr_num}{$el1}{1}}, $pop_frqs{$chr_num}{$el1}{$sbp}{'ALT_ALLELE'};
	push @{$frq_template{$chr_num}{$el1}{1}}, $pop_frqs{$chr_num}{$el1}{$sbp}{'ALT_ALLELE_FRQ'};
	} 
	} 


	elsif (exists $frq_template{$chr_num}{$el1}{1} && !defined $pop_frqs{$chr_num}{$el1}{$sbp}) {
	if ($sample_frqs{$chr_num}{$el1}{'al1'} eq @{$frq_template{$chr_num}{$el1}{1}}[0] && $sample_frqs{$chr_num}{$el1}{'al1_frq'} >= 0.5) {
	push @{$frq_template{$chr_num}{$el1}{1}}, $sample_frqs{$chr_num}{$el1}{'al1'}, 0.99; 
	}
	elsif ($sample_frqs{$chr_num}{$el1}{'al1'} eq @{$frq_template{$chr_num}{$el1}{1}}[0] && $sample_frqs{$chr_num}{$el1}{'al1_frq'} < 0.5) {
	push @{$frq_template{$chr_num}{$el1}{1}}, $sample_frqs{$chr_num}{$el1}{'al1'}, 0.01; 
	}
	elsif ($sample_frqs{$chr_num}{$el1}{'al2'} eq @{$frq_template{$chr_num}{$el1}{1}}[0] && $sample_frqs{$chr_num}{$el1}{'al2_frq'} >= 0.5) {
	push @{$frq_template{$chr_num}{$el1}{1}}, $sample_frqs{$chr_num}{$el1}{'al2'}, 0.99; 
	}
	elsif ($sample_frqs{$chr_num}{$el1}{'al2'} eq @{$frq_template{$chr_num}{$el1}{1}}[0] && $sample_frqs{$chr_num}{$el1}{'al2_frq'} < 0.5) {
	push @{$frq_template{$chr_num}{$el1}{1}}, $sample_frqs{$chr_num}{$el1}{'al2'}, 0.01; 
	}
	}

	}


	}
	}

return \%frq_template, \@subpop_frq_template;
}


sub adms_local {

	my ($parameters_ref, $adm_ref, $bp_pos_ref, $chr_ref, $family_ref, $extrapolation_ref) = @_; # $extrapolation = 0 (yes, default value) or 1 (no extrapolation) 
	my %parameters = %$parameters_ref;
	my %adm = %$adm_ref;
	my $chr = $$chr_ref;
	my $family = $$family_ref;
	my @bp_pos = @$bp_pos_ref;
	@bp_pos =  sort @bp_pos;
	my $extrapolation = $$extrapolation_ref;

	my @aoa =();
	my %local_adms = ();

	for my $key2 (sort {$a <=> $b} keys %{ $adm{$chr} }) { # bp_position
	my @row_aoa =();
	push @row_aoa, $key2;
	for my $key3 (sort {$a cmp $b} keys %{ $adm{$chr}{$key2} }) { # subpopulation
	push @row_aoa, $adm{$chr}{$key2}{$key3}{$family};

	}
	push @aoa, [@row_aoa];

	}


	my @subpops;
 	my %subpop = ();
	my $count1 = 0;
	for my $key2 (sort {$a <=> $b} keys %{ $adm{$chr} }) { # bp_position
	for my $key3 (sort {$a cmp $b} keys %{ $adm{$chr}{$key2} }) { # subpopulation
	push @subpops, $key3;
	$count1++;
	$subpop{$key3} = $count1;
	}
	last;
	}
	my $count2 = scalar @subpops;


	my %adm_blocks = ();
	for my $i (0..($#aoa-1)) {
	for my $j (1..$#{$aoa[$i]}) {
	if ($aoa[$i+1][$j] != $aoa[$i][$j]) {
	push @{$adm_blocks{$aoa[$i][0]}}, (@{ $aoa[$i] } [ 1..($count2)]);
	push @{$adm_blocks{$aoa[$i+1][0]}}, (@{ $aoa[$i+1] } [ 1..($count2)]);
	goto LABEL1;
	}
	}
	LABEL1:
	}


	push @{$adm_blocks{$aoa[0][0]}}, (@{ $aoa[0] } [ 1..($count2)]) if (!defined $adm_blocks{$aoa[0][0]});
	push @{$adm_blocks{$aoa[$#aoa][0]}}, (@{ $aoa[$#aoa] } [ 1..($count2)]) if (!defined $adm_blocks{$aoa[$#aoa][0]});


	my @bp_adm_blocks;
	for my $key ( sort {$a <=> $b} keys %adm_blocks) {
	push @bp_adm_blocks, $key;
	}


	for my $bp (@bp_pos) {
	if (!defined $local_adms{$bp}) {

	if (defined $adm_blocks{$bp}) {
	$local_adms{$bp} = [ @{$adm_blocks{$bp}} ];
	}

	elsif ($bp < $bp_adm_blocks[0] && $extrapolation == 0) { #change if no extrapolation
	$local_adms{$bp} = [ @{$adm_blocks{$bp_adm_blocks[0]}} ];
	}

	elsif ($bp > $bp_adm_blocks[$#bp_adm_blocks] && $extrapolation == 0) { #change if no extrapolation
	$local_adms{$bp} = [ @{$adm_blocks{$bp_adm_blocks[$#bp_adm_blocks]}} ];
	}

	}
	}


	for my $bp (@bp_pos) {
	if (!defined $local_adms{$bp}) {
	for (my $i=0; $i<$#bp_adm_blocks; $i++) {
	if (($bp < $bp_adm_blocks[$i+1]) && ($bp > $bp_adm_blocks[$i])) {
	my $array1_ref = \@{$adm_blocks{$bp_adm_blocks[$i]}};
	my $array2_ref = \@{$adm_blocks{$bp_adm_blocks[$i+1]}};
	my $result1_ref = array_comp ($array1_ref, $array2_ref);
	my $result1 = $$result1_ref;

	if ($result1 == 0) {
	$local_adms{$bp} = [ @{$adm_blocks{$bp_adm_blocks[$i]}} ];
	}

	elsif ($result1 == 1) {
	if ($bp - $bp_adm_blocks[$i] < $bp_adm_blocks[$i+1] - $bp) {
	$local_adms{$bp} = [ @{$adm_blocks{$bp_adm_blocks[$i]}} ];
	}

	elsif ($bp - $bp_adm_blocks[$i] >= $bp_adm_blocks[$i+1] - $bp) { 
	$local_adms{$bp} = [ @{$adm_blocks{$bp_adm_blocks[$i+1]}} ];
	}

	}


	}
	}
	}
	}

	return \(%local_adms, %subpop);
}


sub array_comp { # element-wise comparison of two arrays; input: two arrays to compare, output: 0 = arrays are equal, 1 = array elements are differentat at least at one position

	my ($array_1_ref, $array_2_ref)=@_;
	my @array_1 = @{$array_1_ref};
	my @array_2 = @{$array_2_ref};
	my $result = 0;

	for (my $i=0; $i<=$#array_1; $i++) {
	if ($array_1[$i] == $array_2[$i]) {
	#do nothing
	} else {
	$result = 1;
	}

	} 

	return \$result;
}


sub global_adm_frqs {

	my ($frq_template_ref, $chr_num_ref, $subpop_frq_template_ref, $global_adm_ref) = @_;
	my @subpop_frq_template = @$subpop_frq_template_ref;
	my $chr_num = $$chr_num_ref;
	my %frq_template = %$frq_template_ref;
	my %global_adm = %$global_adm_ref;
	my %global_frqs = ();


	for my $key1 (sort keys %{ $frq_template{$chr_num} }) { # key1 = bp position
	push @{$global_frqs{$chr_num}{$key1}{1}}, $frq_template{$chr_num}{$key1}{1}[0];
	push @{$global_frqs{$chr_num}{$key1}{2}}, $frq_template{$chr_num}{$key1}{2};

	my @temp1;
	for (my $i=1; $i<=$#subpop_frq_template+1; $i++) {
	my $temp2 = ($i-1) + $i;
	my $temp3 = $frq_template{$chr_num}{$key1}{1}[$temp2] * $global_adm{$subpop_frq_template[$i-1]};
	push @temp1, $temp3;
	}
	my $temp4;
	for (@temp1) {
	$temp4 += $_;
	}
	$temp4 = sprintf ("%.6f", $temp4);
	push @{$global_frqs{$chr_num}{$key1}{1}}, $temp4 if ($temp4 < 1);
	push @{$global_frqs{$chr_num}{$key1}{1}}, 0.999999 if ($temp4 >= 1 && $temp4 <= 1.02);
	print "warning\n" if ($temp4 > 1.02);# add to error sub

	if ($global_frqs{$chr_num}{$key1}{1}[1] == 0.999999) {
	push @{$global_frqs{$chr_num}{$key1}{2}}, 0.000001;
	} else {
	my $temp5 = 1 - $temp4;
	$temp5 = sprintf ("%.6f", $temp5) ;
	push @{$global_frqs{$chr_num}{$key1}{2}}, $temp5;
	}

	}

	return \%global_frqs;
}


sub fbgw_frqs {

	my ($frq_template_ref, $subpop_frq_template_ref, $fbgw_ref, $family_ref, $chr_num_ref) = @_;

	my %frq_template = %$frq_template_ref;
	my @subpop_frq_template = @$subpop_frq_template_ref;
	my %fbgw = %$fbgw_ref;
	my $family = $$family_ref;
	my $chr_num = $$chr_num_ref;

	my %fbgw_frqs = ();


	for my $key1 (sort keys %{ $frq_template{$chr_num} }) { # key1 = bp position
	push @{$fbgw_frqs{$chr_num}{$key1}{1}}, $frq_template{$chr_num}{$key1}{1}[0];
	push @{$fbgw_frqs{$chr_num}{$key1}{2}}, $frq_template{$chr_num}{$key1}{2};
	my @temp1;
	for (my $i=1; $i<=$#subpop_frq_template+1; $i++) {
	my $temp2 = ($i-1) + $i;
	my $temp3 = $frq_template{$chr_num}{$key1}{1}[$temp2] * $fbgw{$family}{$subpop_frq_template[$i-1]};
	push @temp1, $temp3;
	}
	my $temp4;
	for (@temp1) {
	$temp4 += $_;
	}
	$temp4 = sprintf ("%.6f", $temp4);
	push @{$fbgw_frqs{$chr_num}{$key1}{1}}, $temp4 if ($temp4 < 1);
	push @{$fbgw_frqs{$chr_num}{$key1}{1}}, 0.999999 if ($temp4 >= 1 && $temp4 <= 1.02);
	print "warning\n" if ($temp4 > 1.02);

	if ($fbgw_frqs{$chr_num}{$key1}{1}[1] == 0.999999) {
	push @{$fbgw_frqs{$chr_num}{$key1}{2}}, 0.000001;
	} else {
	my $temp5 = 1 - $temp4;
	$temp5 = sprintf ("%.6f", $temp5) ;
	push @{$fbgw_frqs{$chr_num}{$key1}{2}}, $temp5;
	}

	}

	return \%fbgw_frqs;
}


sub fbcw_frqs {

	my ($frq_template_ref, $subpop_frq_template_ref, $fbcw_ref,  $family_ref, $chr_num_ref) = @_;

	my %frq_template = %$frq_template_ref;
	my @subpop_frq_template = @$subpop_frq_template_ref;
	my %fbcw = %$fbcw_ref;
	my $family = $$family_ref;
	my $chr_num = $$chr_num_ref;

	my %fbcw_frqs = ();


	for my $key1 (sort keys %{ $frq_template{$chr_num} }) { # key1 = bp position
	push @{$fbcw_frqs{$chr_num}{$key1}{1}}, $frq_template{$chr_num}{$key1}{1}[0];
	push @{$fbcw_frqs{$chr_num}{$key1}{2}}, $frq_template{$chr_num}{$key1}{2};
	my @temp1;
	for (my $i=1; $i<=$#subpop_frq_template+1; $i++) {
	my $temp2 = ($i-1) + $i;
	my $temp3 = $frq_template{$chr_num}{$key1}{1}[$temp2] * $fbcw{$chr_num}{$family}{$subpop_frq_template[$i-1]};
	push @temp1, $temp3;
	}
	my $temp4;
	for (@temp1) {
	$temp4 += $_;
	}
	$temp4 = sprintf ("%.6f", $temp4);
	push @{$fbcw_frqs{$chr_num}{$key1}{1}}, $temp4 if ($temp4 < 1);
	push @{$fbcw_frqs{$chr_num}{$key1}{1}}, 0.999999 if ($temp4 >= 1 && $temp4 <= 1.02);
	print "warning\n" if ($temp4 > 1.02);

	if ($fbcw_frqs{$chr_num}{$key1}{1}[1] == 0.999999) {
	push @{$fbcw_frqs{$chr_num}{$key1}{2}}, 0.000001;
	} else {
	my $temp5 = 1 - $temp4;
	$temp5 = sprintf ("%.6f", $temp5) ;
	push @{$fbcw_frqs{$chr_num}{$key1}{2}}, $temp5;
	}

	}

	return \%fbcw_frqs;
}


sub local_frqs {

	my ($frq_template_ref, $subpop_frq_template_ref, $local_adms_ref,  $family_ref, $chr_num_ref, $subpop_local_ref, $geno_adm_l) = @_; # $geno_adm_l = 0 (no Local adm for geno files) or 1 (Local adm for geno files)

	my %frq_template = %$frq_template_ref;
	my %subpop_local = %$subpop_local_ref;
	my @subpop_frq_template = @$subpop_frq_template_ref;
	my %local_adms = %$local_adms_ref;
	my $family = $$family_ref;
	my $chr_num = $$chr_num_ref;

	my %local_frqs = ();
	my %geno_adm = ();

	for my $key1 (sort {$a <=> $b} keys %{ $frq_template{$chr_num} }) { # key1 = bp position
	push @{$local_frqs{$chr_num}{$key1}{1}}, $frq_template{$chr_num}{$key1}{1}[0];
	push @{$local_frqs{$chr_num}{$key1}{2}}, $frq_template{$chr_num}{$key1}{2};
	my @temp1;
	for (my $i=1; $i<=$#subpop_frq_template+1; $i++) {
	my $temp2 = ($i-1) + $i;
	my $temp3 = $frq_template{$chr_num}{$key1}{1}[$temp2] * $local_adms{$key1}[$subpop_local{$subpop_frq_template[$i-1]}-1];
	$geno_adm{$chr_num}{$key1}{$subpop_frq_template[$i-1]} = sprintf ("%.6f", $local_adms{$key1}[$subpop_local{$subpop_frq_template[$i-1]}-1]) if ($geno_adm_l == 1);


	push @temp1, $temp3;
	}
	my $temp4;
	for (@temp1) {
	$temp4 += $_;
	}
	$temp4 = sprintf ("%.6f", $temp4);
	push @{$local_frqs{$chr_num}{$key1}{1}}, $temp4 if ($temp4 < 1);
	push @{$local_frqs{$chr_num}{$key1}{1}}, 0.999999 if ($temp4 >= 1 && $temp4 <= 1.02);
	print "warning\n" if ($temp4 > 1.02);

	if ($local_frqs{$chr_num}{$key1}{1}[1] == 0.999999) {
	push @{$local_frqs{$chr_num}{$key1}{2}}, 0.000001;
	} else {
	my $temp5 = 1 - $temp4;
	$temp5 = sprintf ("%.6f", $temp5) ;
	push @{$local_frqs{$chr_num}{$key1}{2}}, $temp5;
	}


	}

	return \%local_frqs if ($geno_adm_l == 0);
	return \%local_frqs, \%geno_adm if ($geno_adm_l == 1);
}


sub script_info {

	my $parameters_ref = $_[0];

	my $comp_type = 4;
	my $comp_type_ref = \$comp_type;
	my $dir_ref = make_dir ($parameters_ref, 0, 0, 0, $comp_type_ref);
	my $dir = $$dir_ref;

	my $date_time_2 = localtime;
	$date_time_2 =~ s/\s+/./g;
	$date_time_2 =~ s/\:+/./g;
	my $rand_n = sprintf ("%.0f", rand (999));
	open (LOGFILE, ">", "$dir/admixfrq.$date_time_2.r$rand_n.log");
	my $date_time = localtime;
	print "\n",
	 "  \@----------------------------------------------------------\@\n",
	 "  |    ADMIXFRQ!  |     v1.00      |   19/Jan/2018           |\n",
	 "  |----------------------------------------------------------|\n",
	 "  | (C) 2018 Rafael Nafikov, GNU General Public License, v3  |\n",
	 "  |----------------------------------------------------------|\n",
	 "  | For documentation, citation & bug-reporting instructions:|\n",
	 "  |        https://github.com/RafPrograms/ADMIXFRQ           |\n",
	 "  \@----------------------------------------------------------\@\n",
	 "\n",
	 "	Analysis started at: $date_time\n\n";

	print LOGFILE "\n",
	 "  \@----------------------------------------------------------\@\n",
	 "  |    ADMIXFRQ!  |     v1.00      |   19/Jan/2018           |\n",
	 "  |----------------------------------------------------------|\n",
	 "  | (C) 2018 Rafael Nafikov, GNU General Public License, v3  |\n",
	 "  |----------------------------------------------------------|\n",
	 "  | For documentation, citation & bug-reporting instructions:|\n",
	 "  |        https://github.com/RafPrograms/ADMIXFRQ           |\n",
	 "  \@----------------------------------------------------------\@\n",
	 "\n",
	 "	Analysis started at: $date_time\n\n";

	return *LOGFILE;
}

